from RoboidControl.Thing import Thing
from LinearAlgebra.Spherical import Spherical
from LinearAlgebra.Direction import Direction
from LinearAlgebra.Angle import Angle

class DifferentialDrive(Thing):
    """! A thing which can move itself using a differential drive system
    @sa @link https://en.wikipedia.org/wiki/Differential_wheeled_robot @endlink
    """
    def __init__(self, owner = None, parent = None):
        """! Create a differential drive
        """
        super().__init__(owner = owner, parent = parent)

        ## The radius of a wheel in meters
        self.wheel_radius = 1.0
        ## The distance between the wheels in meters
        self.wheel_separation = 1.0
        
        ## The left wheel
        self.wheel_left: Thing = None
        ## The right wheel
        self.wheel_right: Thing = None

    def SetDriveDimensions(self, wheel_diameter, wheel_separation):
        """! Configures the dimensions of the drive
        @param wheelDiameter The diameter of the wheels in meters
        @param wheelSeparation The distance between the wheels in meters
        
        These values are used to compute the desired wheel speed from the set
        linear and angular velocity.
        @sa SetLinearVelocity SetAngularVelocity
        """
        if wheel_diameter < 0:
            wheel_diameter = -wheel_diameter
        self.wheel_radius = wheel_diameter / 2
        
        if wheel_separation < 0:
            wheel_separation = -wheel_separation
        self.wheel_separation = wheel_separation
        
        self.SetMotors(self.wheel_left, self.wheel_right)

    def SetMotors(self, wheel_left, wheel_right):
        """! Congures the motors for the wheels
        @param leftWheel The motor for the left wheel
        @param rightWheel The motor for the right wheel
        """
        distance: float = self.wheel_separation / 2

        self.wheel_left = wheel_left
        if self.wheel_left is not None:
            self.wheel_left.SetPosition(Spherical(distance, Direction.Left))

        self.wheel_right = wheel_right
        if self.wheel_right is not None:
            self.wheel_right.SetPosition(Spherical(distance, Direction.Right))

    def SetWheelVelocity(self, speed_left, speed_right):
        """! Directly specify the speeds of the motors
        @params speed_left The speed of the left wheel in degrees per second. Positive moves the robot in the forward direction.
        @params speed_right The speed of the right wheel in degrees per second. Positive moves the robot in the forward direction.
        """
        if self.wheel_left is not None:
            self.wheel_left.SetAngularVelocity(Spherical(speed_left, Direction.Left))
        
        if self.wheel_right is not None:
            self.wheel_right.SetAngularVelocity(Spherical(speed_right, Direction.Right))
        
        speed = self.wheel_radius * (speed_left + speed_right) / 2 * Angle.Deg2Rad
        self.SetLinearVelocity(Spherical(speed, Direction.forward))
        steer = self.wheel_radius * (speed_left - speed_right) / self.wheel_separation
        self.SetAngularVelocity(Spherical(steer, Direction.up))

    def Update(self, currentTimeMs, recursive = True):
        """!
        @copydoc Thing::Update
        """
        if not (self.linear_velocity_updated or self.angular_velocity_updated):
            # This assumes no sideward velocity because diff drive does not support it
            linear_speed = self.linear_velocity.distance
            # When the direction is backwards...
            horizontalAngle = self.linear_velocity.direction.horizontal.InDegrees()
            if horizontalAngle < -90 or horizontalAngle > 90:
                linear_speed = -linear_speed

            angular_velocity: Spherical = self.angular_velocity
            angular_speed: float = angular_velocity.distance * Angle.Deg2Rad # in radians/sec
            # Determine the rotation direction
            if angular_velocity.direction.horizontal.InDegrees() < 0:
                angular_speed = -angular_speed
            
            if self.wheel_left is not None and self.wheel_right is not None:
                speed_left: float = 0 if self.wheel_left is None else (linear_speed + angular_speed * self.wheel_left.position.distance) / self.wheel_radius * Angle.Rad2Deg
                speed_right: float = 0 if self.wheel_right is None else (linear_speed - angular_speed * self.wheel_right.position.distance) / self.wheel_radius * Angle.Rad2Deg

                self.SetWheelVelocity(speed_left, speed_right)

        super().Update(currentTimeMs)


